/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * graphics.c: Drawing and sprite management.
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "graphics.h"
#include "board.h"
#include "imath.h"

TEXGROUP(board);

sprite sprites[MAX_SPRITES];
int num_sprites = 0;
objectgroup texgroups;


static struct {
    int posx, posy, posz, ang;

    int vys, vye, vyc;

    bool traced;

    /* Simple one-dimensional depth buffer */
    int depbuf[XDIM];
    int tvx[XDIM];
    int tvy[XDIM];
    int tvx_vy[XDIM];
    int tvy_vx[XDIM];
} gp IBSS_ATTR;

static buffer picbuf;

int loadtex(const char* fname) {
    int i, j, numgroups, result;
    int ptr, dofs;
    texgroup *cg;

    if (picbuf.data) {
        free(picbuf.data);
    }
    for (i = 0; i < texgroups.size; i++) {
        texgroup* tg = (texgroup*)texgroups.objects[i];
        tg->size = 0;
        if (tg->tex)
            free(tg->tex);
    }

    result = loadgzfile(fname, &picbuf);
    if (result < 0) 
	return result;

    ptr = readu32(&picbuf);

#ifdef ROCKBOX_LITTLE_ENDIAN
    picbuf.pos = 4;
    while (picbuf.pos < ptr) {
        int i = readu32(&picbuf);
        *(uint32*)&picbuf.data[picbuf.pos-4] = i;
    }
#endif
    picbuf.pos = ptr;
    numgroups = readu16(&picbuf);

    for (j = 0; j < numgroups; j++) {
	ASSERT(picbuf.pos<picbuf.len, -4);

	cg = (texgroup*) read_namedobject(&texgroups, &picbuf);

	if (cg) {
	    cg->size = readu16(&picbuf);
	    if (!cg->size) continue;
	    cg->tex = malloc(cg->size * sizeof(texture));
	    for (i = 0; i < cg->size; i++) {
		texture *t = &cg->tex[i];
		t->shift = readu8(&picbuf);
                ptr = readu32(&picbuf);
                dofs = readu32(&picbuf);
                
		t->sx = readu32(&picbuf);
		t->ex = readu32(&picbuf);
		t->data = ((uint32*)&picbuf.data[ptr-(dofs<<2)]);
	    }
	} else {
	    uint16 size = readu16(&picbuf);
	    for (i = 0; i < size; i++) {
                picbuf.pos += 17; 
	    }
	}
    }

    

    return 0;
}

int compar_sprite(const sprite * a, const sprite * b) {
    if (a->dist > b->dist)
	return -1;
    if (a->dist < b->dist)
	return 1;

    /* This gurantees a stable sort. */

    if (a < b)
	return -1;
    if (a > b)
	return 1;
    return 0;
}

int pointseen(int x, int y) {
    int bx, by, ox, oy;
    bx = x >> 16;
    by = y >> 16;
    for (ox = -1; ox <= 1; ox++)
	for (oy = -1; oy <= 1; oy++)
	    if (!gp.traced || BSEEN(bx + ox, by + oy))
                return 1;
    return 0;

}

void addsprite_tex(int x, int y, int z, texture* tex, spritedir dir, int flags) {
    int bx, by, ox, oy;
    sprite *s;
    if (num_sprites == MAX_SPRITES)
	return;
    if (pointseen(x, y)) {
        s = &sprites[num_sprites++];
        s->x = x;
        s->y = y;
        s->z = z;
        s->tex = tex;
        s->flags = flags;
        s->dir = dir;
        s->dist = 0;
    }
}

void _addsprite(int x, int y, int z, texgroup* g, int n, spritedir dir, int flags) {
    if (n < g->size)
	addsprite_tex(x, y, z, &g->tex[n], dir, flags);
}



#if !defined(SDLTEST) && defined(CPU_COLDFIRE)

/* Dark magic happens here. */

#define initprev "move.b (%[p]),%%d0            \n\t" /* 3(1/0) */
#define loadprev "move.b (%[p]),%%d0            \n\t" /* 2(1/0) */
#define initceil                                                        \
         "move.b #0x55, %%d0                    \n\t" /* 1(0/0) */

#define loadfloor                                                       \
         "move.b #0xaa, %%d0                    \n\t" /* 1(0/0) */

#define subone                                                          \
         "subq.l #1, %[y]                       \n\t" /* 1(0/0) */

#define maskone

#define btnoop
#define btskip                                                          \
         "andi.l #7, %[y]                       \n\t" /* 1(0/0) */      \
         "beqs 5f                               \n\t" /* 3(0/0) | 1(0/0) */

#define LOADB                                                           \
	 "move.l %[by], %[y]                    \n\t" /* 1(0/0) */      \
	 "lsr.l %[size], %[y]                   \n\t" /* 1(0/0) */      \
	 "move.l (%[m],%[y]*4),%[y]             \n\t" /* 3(1/0) */      \
	 "lsr.l %[shf], %[y]                    \n\t" /* 1(0/0) */      \
	 "add.l  %[byi], %[by]                  \n\t" /* 1(0/0) */      \


#define NEXTL                                                           \
	 "5:                                    \n\t"                   \
	 "subq.l   #1, %[cnt]                   \n\t" /* 1(0/0) */      \
	 "jbeq     6f                           \n\t" /* 3(0/0) | 1(0/0) */

#define LINELOOP(even,odd,shf,tst,init,load)                            \
    uint8 *__p = p;                                                     \
    int __y = y, __by = by, __byi = byi;                                \
    int __cnt = cnt, __size = size;                                     \
    asm volatile(                                                       \
	 " | LINELOOP                           \n\t"                   \
	 "neg.l %[size]                         \n\t" /* 1(0/0) */      \
         "moveq  #16, %%d0                      \n\t" /* 1(0/0) */      \
	 "add.l  %%d0, %[size]                  \n\t" /* 1(0/0) */      \
         "moveq  #3, %%d0                       \n\t" /* 1(0/0) */      \
         "and.l  %%d0, %[y]                     \n\t" /* 1(0/0) */      \
         "eor.l  %%d0, %[y]                     \n\t" /* 1(0/0) */      \
	 init                                                           \
         "jmp.l (2,%%pc,%[y].l*4)               \n\t" /* 4(0/0) */      \
         "bra.w 4f                              \n\t" /* 2(0/0) */      \
         "bra.w 3f                              \n\t" /* 2(0/0) */      \
         "bra.w 2f                              \n\t" /* 2(0/0) */      \
         "                                      \n"                     \
	 "1:                                    \n\t"                   \
	 LOADB tst even                                                 \
	 "                                      \n\t"                   \
	 "lsr.l #1, %[y]                        \n\t" /* 1(0/0) */      \
	 "and.l #3, %[y]                        \n\t" /* 1(0/0) */      \
	 "and.l #0xfc, %%d0                     \n\t" /* 1(0/0) */      \
	 "or.l  %[y],  %%d0                     \n\t" /* 1(0/0) */      \
	 "                                      \n"                     \
	 NEXTL                                                          \
	 "                                      \n"                     \
	 "2:                                    \n\t"                   \
	 LOADB tst odd                                                  \
	 "                                      \n\t"                   \
	 "lsl.l #1, %[y]                        \n\t" /* 1(0/0) */      \
	 "and.l #0x0c, %[y]                     \n\t" /* 1(0/0) */      \
	 "and.l #0xf3, %%d0                     \n\t" /* 1(0/0) */      \
	 "or.l  %[y],  %%d0                     \n\t" /* 1(0/0) */      \
	 "                                      \n"                     \
	 NEXTL                                                          \
	 "                                      \n"                     \
	 "3:                                    \n\t"                   \
	 LOADB tst even                                                 \
	 "                                      \n\t"                   \
	 "lsl.l #3, %[y]                        \n\t" /* 1(0/0) */      \
	 "and.l #0x30, %[y]                     \n\t" /* 1(0/0) */      \
	 "and.l #0xcf, %%d0                     \n\t" /* 1(0/0) */      \
	 "or.l %[y],   %%d0                     \n\t" /* 1(0/0) */      \
	 "                                      \n"                     \
	 NEXTL                                                          \
	 "                                      \n"                     \
	 "4:                                    \n\t"                   \
	 LOADB tst odd                                                  \
	 "                                      \n\t" /* 1(0/0) */      \
	 "lsl.l #5, %[y]                        \n\t" /* 1(0/0) */      \
	 "and.l #0xc0, %[y]                     \n\t" /* 1(0/0) */      \
	 "and.l #0x3f, %%d0                     \n\t" /* 1(0/0) */      \
	 "or.l %[y],   %%d0                     \n\t" /* 1(0/0) */      \
	 "                                      \n"                     \
	 NEXTL                                                          \
	 "                                      \n\t"                   \
	 "move.b %%d0, (%[p])                   \n\t" /* 1(0/1) */      \
	 "adda.l %[bpl], %[p]                   \n\t" /* 1(0/0) */      \
	 load                                                           \
	 "jbra     1b                           \n\t" /* 2(0/0) */      \
	 "                                      \n"                     \
	 "6:                                    \n\t"                   \
	 "move.b %%d0, (%[p])                   \n"   /* 1(0/1) */      \
	 : [p] "+a" (__p), [cnt] "+d" (__cnt),                          \
           [size] "+d" (__size), [by] "+d" (__by),                      \
	   [byi] "+d" (__byi), [y] "+d" (__y)                           \
	 : [m] "a" (m), [bpl] "i" (BPL), [shf] "d" (shf)                \
	 : "d0", "memory" );

#else /* CPU_COLDFIRE */

#define subone v--
#define maskone
#define btnoop if (1) 
#define btskip if (v&7)

#define initceil 0x55
#define loadfloor 0xaa
#define initprev *__p
#define loadprev *__p

#define LOADB v = ((m[__by>>__size])>>shf)&7; __by += byi
#define NEXTL if (!--__cnt) break
#define LINELOOP(even,odd,shf,tst,init,load)                                    \
    uint8* __p = p;                                                             \
    uint32 v, s;                                                                \
    int __by = by;                                                              \
    int __cnt = cnt, __size = 16-size;                                          \
    s = init;                                                                   \
    /* Duff's device: */                                                        \
    switch (y&3) {                                                              \
	do {                                                                    \
	case 0: LOADB; tst { even; s &= 0xfc; s |= (v>>1)&0x03; } NEXTL;        \
	case 1: LOADB; tst { odd;  s &= 0xf3; s |= (v<<1)&0x0c; } NEXTL;        \
	case 2: LOADB; tst { even; s &= 0xcf; s |= (v<<3)&0x30; } NEXTL;        \
	case 3: LOADB; tst { odd;  s &= 0x3f; s |= (v<<5)&0xc0; } NEXTL;        \
	    *__p = s; __p += BPL; s = load;                                     \
	} while (1);                                                            \
    }                                                                           \
    *__p = s;




#endif /* CPU_COLDFIRE */

unsigned short divmod10x3[512] IBSS_ATTR;

static CTOR void __init_divtbl(void) {
    int i;
    int m = 0, d = 0;
    unsigned short *p = divmod10x3;
    for (i = 0; i < 512; i++, p++) {
        *p = (d<<8) | m;
        m += 3;
        if (m == 30) {
            m = 0;
            d++;
        }
    }
}

#define VERTLINE(tst,in,ld)                             \
    if (cnt) {                                          \
        int col = bx>>(16-size);                        \
        int shf = divmod10x3[col];                      \
	m = &data[(shf>>8)<<size];                      \
        shf &= 255;                                     \
	if (sx&1) {                                     \
	    LINELOOP(subone,maskone,shf,tst,in,ld);     \
	} else {                                        \
	    LINELOOP(maskone,subone,shf,tst,in,ld);     \
	}                                               \
    }


void preparescreen(int posx, int posy, int posz, int ang, int vys, int vye, int vyc) {
    int sinang, cosang;
    int vxinc, vyinc, vx, vy, sx, vx_vy, vy_vx;

    gp.posx = posx;
    gp.posy = posy;
    gp.posz = posz;
    gp.vys = vys;
    gp.vye = vye;
    gp.vyc = vyc;
    gp.ang = ang;
    sinang = isin(ang);
    cosang = icos(ang);

    vxinc = -divconst(sinang << 9,FOVMUL);
    vx = (cosang - (vxinc>>1));

    vyinc = divconst(cosang << 9,FOVMUL);
    vy = (sinang - (vyinc>>1));

    vxinc = divconst(vxinc,XDIM);
    vyinc = divconst(vyinc,XDIM);

    gp.traced = false;

    for (sx = 0; sx < XDIM; sx++) {
        gp.depbuf[sx] = 0x7fffffff;

	vx_vy = div32(vx<<14, vy>>2);
	vy_vx = div32(vy<<14, vx>>2);

	gp.tvx[sx] = vx; gp.tvy[sx] = vy;
	gp.tvx_vy[sx] = vx_vy;
	gp.tvy_vx[sx] = vy_vx;

	vx += vxinc;
	vy += vyinc;
    }
}

void drawsprites(uint8 * screen) {
    int sinang, cosang;
    int xdist;
    int hx, hy;
    int dist;
    int sx, seen;
    int tbx = 0, bx = 0, by = 0, byi = 0, sby;
    int i;
    int y, ye, cnt;
    int size;
    uint8 *p, *base;
    uint32 *m;
    uint32 *data;
    int sxs, sxe;


    sprite *cs;
    sinang = isin(gp.ang);
    cosang = icos(gp.ang);

    cs = sprites;
    for (i = 0; i < num_sprites; i++, cs++) {
	cs->dist = dotprod(cosang, sinang, cs->x - gp.posx, cs->y - gp.posy);
        switch (cs->dir) {
        case DIR_WEST:
            if (gp.posx >= cs->x) {
                cs->dir = DIR_HIDE;
                continue;
            }
            cs->y += 32768;
            break;
        case DIR_EAST:
            if (gp.posx <= cs->x) {
                cs->dir = DIR_HIDE;
                continue;
            }
            cs->y -= 32768;
            break;
        case DIR_NORTH:
            if (gp.posy >= cs->y) {
                cs->dir = DIR_HIDE;
                continue;
            }
            cs->x += 32768;
            break;
        case DIR_SOUTH:
            if (gp.posy <= cs->y) {
                cs->dir = DIR_HIDE;
                continue;
            }
            cs->x -= 32768;
            break;
        default:
            break;
        }
    }
    RB qsort(sprites, num_sprites, sizeof(sprite),
             (int (*)(const void *, const void *))compar_sprite);

    cs = sprites;
    for (i = 0; i < num_sprites; i++, cs++) {
        /* if we're drawing a lot of sprites, give the sound thread
           a chance to run */
#ifndef SDLTEST
        if (i && ((i&3) == 0)) RB yield();
#endif

        spritedir dir = cs->dir;
        int flip = (cs->flags & SF_FLIPX) != 0;
        int xp, less;

        switch (dir) {
        case DIR_HIDE:
            continue;
        case DIR_WEST:
            xp = 0;
            less = 1;
            flip ^= 1;
            break;
        case DIR_EAST:
            xp = 0;
            less = 0;
            flip ^= 1;
            break;
        case DIR_NORTH:
            xp = 1;
            less = 1;
            break;
        case DIR_SOUTH:
            xp = 1;
            less = 0;
            break;
        case DIR_FACE:
            if (cs->dist <= 1024)
                continue;
        default:
            break;
        }
	texture* tex = cs->tex;

	xdist = dotprod(-sinang, cosang, cs->x - gp.posx, cs->y - gp.posy);
        if (flip) {
            sxs = 65536-tex->ex;
            sxe = 65536-tex->sx;
        } else {
            sxs = tex->sx;
            sxe = tex->ex;
        }
	size = tex->shift;
	data = tex->data;

	if (dir != DIR_FACE) {
            sx = 0;
	    seen = 0;
	    sby = (gp.posz - cs->z + 32768);
	    for (sx = 0; sx < XDIM; sx++) {
		if (xp) {
		    if (!gp.tvy[sx])
			break;
		    hy = cs->y;
		    hx = gp.posx + fixedmul(gp.tvx_vy[sx],hy - gp.posy);
		    bx = (hx - cs->x);
		    if (less)
			bx = -bx;
		    if (bx < sxs || bx >= sxe || ((gp.tvy[sx] > 0) != less)) {
			if (seen)
			    break;
			else
			    continue;
		    }
		} else {
		    if (!gp.tvx[sx])
			break;
		    hx = cs->x;
		    hy = gp.posy + fixedmul(gp.tvy_vx[sx],hx - gp.posx);
		    bx = (hy - cs->y);
		    if (less)
			bx = -bx;
		    if (bx < sxs || bx >= sxe || ((gp.tvx[sx] > 0) != less)) {
			if (seen)
			    break;
			else
			    continue;
		    }
		}
                dist = dotprod(cosang, sinang, hx - gp.posx, hy - gp.posy);
                byi = divconst(dist,HVAL);
                if (flip)
                    bx = 65536-bx;

		seen = 1;

		if (dist > gp.depbuf[sx] || dist <= 256)
		    continue;


		by = sby;
		y = -(by/byi);
		ye = gp.vyc + (65536-by)/byi;
		if (y < gp.vys-gp.vyc) {
		    by += (gp.vys-gp.vyc) *byi;
		    y = gp.vys;
		} else {
		    by += y*byi;
		    y += gp.vyc;
		}
                if (sby < 0) {
                    sby += byi;
                }
		if (ye > gp.vye) ye = gp.vye;
		cnt = ye-y;
                if (cnt <= 0) continue;
		
		p = &screen[sx + (BPL * (y>>2))];

		VERTLINE(btskip,initprev,loadprev);
	    
	    }
	} else {
	    dist = cs->dist;
	    if (dist <= 256)
		continue;
	    sx = (XDIM/2) + mul64(xdist - 32768 + sxs, HVAL ) / dist;
	    if (sx < 0)
		sx = 0;
	    byi = divconst(dist,HVAL);
	    bx = (sx + 1 - XDIM/2) * byi + ((-xdist + 32768)) - 1;

	    sby = gp.posz - cs->z + 32768;
	    y = -(sby/byi);
	    ye = gp.vyc + (65536-sby)/byi;
	    if (y < gp.vys-gp.vyc) {
		sby += (gp.vys-gp.vyc)*byi;
		y = gp.vys;
	    } else {
		sby += y*byi;
		y += gp.vyc;
	    }
            if (sby < 0) {
                sby += byi;
                y++;
            }
	    if (ye > gp.vye) ye = gp.vye;

            tbx = bx;
	    by = sby;
	    size = tex->shift;
	    cnt = ye-y;
            if (cnt <= 0) continue;
	    data = tex->data;
	    base = &screen[sx + (BPL * (y>>2))];
	    for (; sx < XDIM && tbx < sxe; sx++, tbx += byi, base++) {
		if (dist > gp.depbuf[sx])
		    continue;
                if (flip)
                    bx = 65536-tbx;
                else
                    bx = tbx;
		p = base;
		VERTLINE(btskip,initprev,loadprev);
	    }
	}
    }
    
    num_sprites = 0;

}
void drawscreen(uint8 * screen) {
    int sinang, cosang, xtemp, ytemp;
    int vxinc, vyinc, vx, vy, sx, vx_vy, vy_vx;
    int xscan, yscan, xdir, ydir, incx, incy;
    int d, hx, hy;

    int bx, by, byi, sby = gp.posz + 32768;
    int dist;
    int y, ye, cnt;
    int size = 9;
    int pn;
    uint8 *p; 
    uint32 *m;
    uint32 *data;


    RB memset(boardseen, 0, 4096);

    sinang = isin(gp.ang);
    cosang = icos(gp.ang);

    gp.traced = true;

    for (sx = 0; sx < XDIM; sx++) {
	vx = gp.tvx[sx]; vy = gp.tvy[sx];
	vx_vy = gp.tvx_vy[sx]; vy_vx = gp.tvy_vx[sx];

	xscan = gp.posx >> 16;
	xdir = vx > 0 ? 1 : -1;
	incx = abs(vx);

	yscan = gp.posy >> 16;
	ydir = vy > 0 ? 1 : -1;
	incy = abs(vy);

	xtemp = gp.posx & 65535;
	if (xdir > 0)
	    xtemp = 65536 - xtemp;
	ytemp = gp.posy & 65535;
	if (ydir > 0)
	    ytemp = 65536 - ytemp;

	d = dotprod(xtemp, ytemp, incy, -incx);

	while (1) {
            /* safety */
	    if (xscan < 0 || xscan > 63 || yscan < 0 || yscan > 63) {
		return;
	    }
            BSEEN(xscan, yscan) = 1;
	    if (d < 0) {
		pn = BX(xdir, xscan, yscan) & 4095;
		xscan += xdir;
		if (pn) {
		    hx = xscan << 16;
		    if (xdir < 0)
			hx += 65536;
		    hy = gp.posy + fixedmul(vy_vx,hx - gp.posx);
		    bx = (hy & 65535);
		    if (xdir < 0) {
			bx = 65535 - bx;
		    }
		    break;
		}
		d += incy;
	    } else {
		pn = BY(ydir, xscan, yscan) & 4095;
		yscan += ydir;
		if (pn) {
		    hy = yscan << 16;
		    if (ydir < 0)
			hy += 65536;
		    hx = gp.posx + fixedmul(vx_vy,hy - gp.posy);
		    bx = (hx & 65535);
		    if (ydir > 0) {
			bx = 65535 - bx;
		    }
		    break;
		}
		d -= incx;
	    }
	}

	dist = gp.depbuf[sx] = dotprod(cosang, sinang, hx - gp.posx, hy - gp.posy);

	if (dist < 1024)
	    continue;

	byi = divconst(dist,HVAL);

	if (pn >= texgrp_board.size)
	    pn = 0;

	by = sby;
	y = -div32(by,byi);
	ye = gp.vyc + div32((65536-by),byi);
	if (y < gp.vys-gp.vyc) {
	    by += (gp.vys-gp.vyc)*byi;
	    y = gp.vys;
	} else {
	    by += y*byi;
	    y += gp.vyc;
        }
	if (ye > gp.vye) ye = gp.vye;
	cnt = ye-y;
        if (cnt <= 0) continue;
	if (y == gp.vyc) y = gp.vyc-1; /* HACK! */
	p = &screen[sx + (BPL * (y>>2))];
	size = texgrp_board.tex[pn].shift;
	data = texgrp_board.tex[pn].data;
	VERTLINE(btnoop,initceil,loadfloor);
    }
}
