/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * player.c: Manages player and player bullet
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "player.h"
#include "graphics.h"
#include "board.h"
#include "soundutil.h"
#include "miscactors.h"
#include "imath.h"

TEXGROUP(bullet);
ACTOR_ICLASSDEF(playerbullet, int ttl; int sdir);
ACTOR_ICLASSDEF(player,
	  int fvel, avel, svel, zvel;
	  int posz;
	  int fire, firetime;
	  int hurttime, health;
	  int deathtime;
	  int zbase);

METHOD(void, playerbullet, init) {
    actor_init(self);

    self->ttl = 24;
};

METHOD(void, player, init) {
    actor_init(self);

    self->fvel = self->avel = self->svel = self->zvel = 0;
    self->posz = self->firetime = self->fire = 0;
    self->hurttime = 0;
    self->health = 10;
    self->w = self->h = 8192;
    self->deathtime = 0;
    self->zbase = 0;
};

SERIALIZER(player) {
    S_INT(posz);
    S_INT(firetime);
    S_INT(hurttime);
    S_INT(health);
    S_INT(deathtime);
    S_INT(zbase);
}

SERIALIZER(playerbullet) {
    S_INT(ttl);
    S_INT(sdir);
}

METHOD(void, playerbullet, move) {
    if (!self->ttl--) {
	actor_remove(self);
	return;
    }

    int res = actor_move(self);
    if (res & 1)
	self->vx = -self->vx;
    if (res & 2)
	self->vy = -self->vy;
    if (res)
	playsndxy("bounce", self->x, self->y, 64);
}

METHOD(void, playerbullet, touch, actor* o) {
    if (!HASTAG(self, actor_remove)) {
	int res = CALLMETH(o, damage, TAGSET(damages_monster, damages_misc), 1);
	if (res != DMG_NO) {
	    actor_remove(self);
	    if (res != DMG_ABSORB) {
		addexplosion(self->x, self->y);
		playsndxy("explo", self->x, self->y, 64);
	    }
	}
    }
}


METHOD(void, playerbullet, draw) {
    int xttl = 24 - self->ttl;
    int a = (xttl * 280 * self->sdir);
    int t2 = (xttl * (xttl - 24)) * 3;
    int s = (isin(a) * t2)>>12;
    int c = (icos(a) * t2)>>12;
    int s2 = isin(self->ang);
    int c2 = icos(self->ang);
    int tx = self->x + fixedmul(s2, c);
    int ty = self->y - fixedmul(c2, c);
    if (!HASTAG(self, actor_remove)) {
	addsprite(tx, ty, s, bullet, 0, DIR_FACE, 0);
    }
}



METHOD(void, player, move) {
    int sinang = isin(self->ang);
    int cosang = icos(self->ang);
    int dt2;
    static int bulsdir = 1;
    if (self->hurttime)
	self->hurttime--;

    if (self->deathtime >= 30) {
        ADDTAG(self, is_dead);
        return;
    }
    if (self->deathtime) {
	self->deathtime++;
	dt2 = self->deathtime * self->deathtime;
	self->posz = self->zbase + ((dt2 - 16 * self->deathtime) << 8);
	self->ang += 112;
	if (self->posz < -30720)
	    self->posz = -30720;
	if (self->posz > 30720)
	    self->posz = 30720;
	return;
    }
    self->ang += self->avel * 48;
    self->posz += self->zvel << 11;

    self->vx = -self->svel * ((sinang * 3) >> 4) + self->fvel * ((cosang * 3) >> 4);
    self->vy = self->svel * ((cosang * 3) >> 4) + self->fvel * ((sinang * 3) >> 4);

    if (self->posz < -30720)
	self->posz = -30720;
    if (self->posz > 30720)
	self->posz = 30720;

    if (self->firetime)
	self->firetime--;

    if (self->fire && !self->firetime) {
	playerbullet *o = object_new(playerbullet);
        
	if (!o)
	    return;
        o->x = self->x;
        o->y = self->y;
        o->ang = self->ang;
	o->vy = sinang >> 1;
	o->vx = cosang >> 1;
	o->w = o->h = 8192;
        o->sdir = bulsdir;
        bulsdir *= -1;
	if (!actor_add((actor *) o)) {
            object_deref(o);
	}
	playsndxy("player_shoot", self->x, self->y, 64);
	self->firetime = 6;
    }
    actor_move(self);
}

METHOD(int, player, damage, tagset* type, int amt) {
    if (!HASTAG(self, is_invulnerable) && TAGSET_IN(type, damages_player)) {
	if (!self->hurttime && self->health) {
	    self->hurttime = 18;
	    self->health -= amt;
	    if (self->health <= 0) {
                ADDTAG(self, is_dying);
		self->health = 0;
		playsndxy("player_die", self->x, self->y, 64);
		self->deathtime = 1;
		self->zbase = self->posz;
	    } else {
		playsndxy("player_hurt", self->x, self->y, 64);
	    }
	}
	return DMG_YES;
    }
    return DMG_NO;
}


METHOD(void, player, load, buffer* b) {
    actor_load(self, b);
    ADDTAG(self, is_player);
}

int player_get_health(actor* o) {
    return ((player*)o)->health;
}

void player_set_controls(actor* o, int f, int s, int z, int a, int fire) {
    player* self = (player*)o;

    self->fvel = f;
    self->svel = s;
    self->zvel = z;
    self->avel = a;
    self->fire = fire;
}
int player_get_posz(actor* o) {
    return ((player*)o)->posz;
}
int player_get_hurttime(actor* o) {
    return ((player*)o)->hurttime;
}
int player_add_health(actor* o, int h) {
    if (((player*)o)->health >= 10) return 0;
    ((player*)o)->health += h;
    if (((player*)o)->health > 10) ((player*)o)->health = 10;
    return 1;
}

void player_copy(actor* from, actor* to) {
    player* o = (player*)from;
    player* self = (player*)to;

    tagset_copy(&self->tags, &o->tags);
    DELTAG(self, has_key);
    
    self->deathtime = 0;
    self->fvel = self->avel = self->svel = self->zvel = 0;
    self->posz = self->firetime = self->fire = 0;
    self->hurttime = 0;
    self->health = o->health;
}
