/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * tag.c: Manages tags
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "tag.h"

#define isdigit(c) ((c) >= '0' && (c) <= '9')
#define isalpha(c) (((c) >= 'a' && (c) <= 'z') || ((c) >= 'A' && ((c) <= 'Z')))
#define isalnum(c) (isdigit(c) || isalpha(c))
#define isident(c) (isalnum(c) || ((c) == '_'))

static int ntags = 0;
static int tagcap = 16;

INITVAR(static const char**, tagtbl, malloc(sizeof(char*) * tagcap));

tag _get_tag(const char* name, int dup) {
    int i;
    const char** ctag = tagtbl;
    for (i = 0; i < ntags; i++, ctag++) {
        if (!RB strcmp(*ctag, name)) {
            //printf("get tag %d = %s\n", i+1, name);
            return i+1;
        }
    }
    if (dup) {
        name = strdup(name);
    }
    expand(tagtbl, tagcap, ntags + 1);
    tagtbl[ntags++] = name;
    //printf("new tag %d = %s\n", ntags, name);
    return ntags;
}

const char* tag_name(tag t) {
    return tagtbl[t - 1];
}

void convert_long(tagset* ts, uint32 len) {
    uint32* ldata = calloc(sizeof(long), len);
    RB memcpy(ldata, ts, sizeof(long) * 4);

    ts->flags = 1;
    ts->len = len;
    ts->data = ldata;
}

void convert_short(tagset* ts) {
    uint32* ldata = ts->data;

    RB memcpy(ts, ldata, sizeof(long) * 4);
    ts->flags &= ~1;

    free(ldata);
}

int _tagset_in_long(const tagset* ts, tag t) {
    uint32 idx = t>>5;
    if (idx >= ts->len)
        return 0;
    
    return (ts->data[idx] >> (t&31)) & 1;
}

void _tagset_copy_long(tagset* to, const tagset* from) {
    tagset_clear(to);
    uint32* ldata = calloc(sizeof(long), from->len);
    to->len = from->len;
    to->data = ldata;
    to->flags = 1;

    RB memcpy(to->data, from->data, sizeof(long) * from->len);
}

void _tagset_clear_long(tagset* ts) {
    free(ts->data);
    ts->shdata[0] = ts->shdata[1] = 
        ts->shdata[2] = ts->shdata[3] = 0;
}

void _tagset_add_long(tagset* ts, tag t) {
    uint32 idx = t>>5;
    uint32 i;

    if (!(ts->flags & 1)) {
        convert_long(ts, idx + 1);
    }
    if (idx >= ts->len) {
        ts->data = realloc(ts->data, idx + 1);
        for (i = ts->len; i <= idx; i++) {
            ts->data[i] = 0;
        }
        ts->len = idx + 1;
    }
    ts->data[idx] |= 1<<(t&31);
}

void _tagset_del_long(tagset* ts, tag t) {
    uint32 idx = t>>5;
    if (idx >= ts->len)
        return;

    ts->data[idx] &= ~(1<<(t&31));

    uint32 i = ts->len - 1;

    while (i > 3) {
        if (ts->data[i])
            break;
        i--;
    }

    if (i == 3) {
        convert_short(ts);
        return;
    }
}



void _tagset_compose(tagset* ts, const char* lst) {
    const char* end;
    while (1) {
        while (*lst && !isident(*lst)) lst++;
        if (!*lst) return;
        end = lst;
        while (*end && isident(*end)) end++;
        char tmp[(end - lst) + 1];
        RB memcpy(tmp, lst, (end - lst));
        tmp[end - lst] = 0;
        tag tt = _get_tag(tmp, 1);
        tagset_add(ts, tt);
        lst = end;
    }
}


void tagset_enumerate(const tagset* ts, void (*func)(tag, void*), void* cookie) {
    uint32 len = 4;
    const uint32* data = ts->shdata;
    uint32 i, j;

    if (ts->flags & 1) {
        len = ts->len;
        data = ts->data;
    }
    for (i = 0; i < len; i++) {
        for (j = 0; j < 32; j++) {
            if (data[i] & (1<<j)) {
                func((i<<5) | j, cookie);
            }
        }
    }
}

static void _tagset_count(tag t, uint32* count) {
    (void)t;
    (*count)++;
}

static void _tagset_list_add(tag t, tag** ct) {
    *(*ct)++ = t;
}

void tagset_list(const tagset* ts, tag** tags, uint32* ntags) {
    uint32 nt = 0;
    tag* ct;
    tagset_enumerate(ts, (void(*)(tag, void*))_tagset_count, &nt);
    ct = (tag*)malloc(sizeof(tag) * nt);
    *tags = ct;
    *ntags = nt;
    tagset_enumerate(ts, (void(*)(tag, void*))_tagset_list_add, &ct);
}

static void _tagset_write(tag t, buffer* buf) {
    writestr(buf, tag_name(t));
}

void tagset_write(buffer* buf, tagset* ts) {
    uint32 nt = 0;
    tagset_enumerate(ts, (void(*)(tag, void*))_tagset_count, &nt);
    writeu16(buf, nt);
    tagset_enumerate(ts, (void(*)(tag, void*))_tagset_write, buf);
}
void tagset_read(buffer* buf, tagset* ts) {
    tagset_init(ts);
    uint32 nt = readu16(buf);
    uint32 i;
    char tmp[64];

    for (i = 0; i < nt; i++) {
        readstr(buf, tmp, 64);
        tagset_add(ts, _get_tag(tmp, 1));
    }
}



void serial_tagset(tagset* ts, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        tagset_write(buf, ts);
    } else if (phase == PHASE_READ) {
        tagset_read(buf, ts);
    }
}
